import os
import time
import orjson
import shutil
import logging

from fvs.exceptions import FVSNothingToCommit, FVSEmptyCommitMessage, FVSStateNotFound, FVSMissingStateIndex, \
    FVSNothingToRestore, FVSStateZeroNotDeletable, FVSEmptyStateIndex, FVSStateAlreadyExists
from fvs.pattern import FVSPattern
from fvs.state import FVSState
from fvs.file import FVSFile
from fvs.data import FVSData
from fvs.utils import FVSUtils

logger = logging.getLogger("fvs.repo")


class FVSRepo:
    __repo_conf: dict = None
    __has_no_states: bool = False
    __use_compression = False
    __active_state: 'FVSState' = None

    def __init__(self, repo_path: str, use_compression: bool = False, no_init: bool = False):
        """
        Initialize the FVSRepo.
        """
        self.__repo_path = os.path.abspath(repo_path)
        self.__states_path = os.path.join(self.__repo_path, ".fvs/states")
        self.__use_compression = use_compression
        if not no_init:
            self.__update_fvs_path()
        self.__load_config()

    def __update_fvs_path(self):
        """
        Update the path of the .fvs directory. This directory is not meant
        to be edited outside/manually. Here we just create it if it doesn't
        exist and try to fix it if wrong.
        """
        dirs = [
            ".fvs",
            ".fvs/states",
            ".fvs/data",
        ]
        repo_conf = os.path.join(self.__repo_path, ".fvs/repo.json")
        updated = False
        for _dir in dirs:
            if not os.path.exists(os.path.join(self.__repo_path, _dir)):
                os.makedirs(os.path.join(self.__repo_path, _dir))
                updated = True

        if not os.path.exists(repo_conf):
            with open(repo_conf, "wb") as f:
                self.__repo_conf = {"id": -1, "states": {}, "compression": self.__use_compression}
                f.write(orjson.dumps(self.__repo_conf, f, option=orjson.OPT_NON_STR_KEYS,))
                updated = True

        if updated:
            logger.debug(f"FVS path updated for repository {self.__repo_path}")

    def __load_config(self):
        """
        Load the repository configuration.
        """
        repo_conf = os.path.join(self.__repo_path, ".fvs/repo.json")
        with open(repo_conf, "r") as f:
            self.__repo_conf = orjson.loads(f.read())

        """
        JSON store int key as strings, so we need to convert them back to int.
        """
        self.__states = {int(key): value for key, value in self.__repo_conf["states"].items()}

        if int(self.__repo_conf["id"]) >= 0:
            self.__active_state = FVSState(self, int(self.__repo_conf["id"]))
        else:
            self.__has_no_states = True
        
        self.__use_compression = self.__repo_conf["compression"]

    def get_unstaged_files(self, ignore: list = None, purpose: int = 0) -> dict:
        """
        Get the unstaged files.
        ...
        Purpose values:
            0: Committing a new state.
            1: Restoring a state (will return original sha1 for modified files)
        """
        unstaged_files = {
            "count": 0,
            "added": [],
            "removed": [],
            "modified": [],
            "intact": []
        }
        active_state_files = {}

        if ignore is None:
            ignore = []

        """
        The following new variable is used to store all relative paths
        handled in the following loop. We will use them to list removed
        files.
        """
        unstaged_relative_paths = []

        """
        Create a copy of the active state files so we can remove every
        handled entry and easily figure out what was deleted.
        """
        if not self.__has_no_states:
            active_state_files = self.__active_state.files.copy()
        else:
            active_state_files = {"added": {}, "removed": {}, "modified": {}, "intact": {}}

        def del_active_state_file_key(sha1: str):
            active_state_files["added"].pop(sha1, None)
            active_state_files["modified"].pop(sha1, None)
            active_state_files["intact"].pop(sha1, None)
        
        for root, _, files in os.walk(self.__repo_path):
            """
            Here we are excluding the .fvs/ directory from the unstaged files
            because we don't want to invoke the monster of loops.
            """
            if ".fvs" in root.split(os.sep):
                continue

            """
            Here we loop through the files and determinate which ones are
            added, removed, modified or intact, comparing with prior state 
            or simply adding all of them if this is the first state.
            """
            for file in files:
                _full_path = os.path.join(root, file)
                _relative_path = self.__get_relative_path(os.path.join(root, file))

                """
                Here we loop through the ignore pattern and remove the files that
                match any of them. Check if performed on the relative path.
                """
                if FVSPattern.match(ignore, _relative_path):
                    continue

                """
                Here we calculate the sha1 of the file. It will return None if
                the file is not accessible or doesn't exist.
                """
                _sha1 = FVSUtils.get_sha1_hash(_full_path)
                if _sha1 is None:
                    continue
                
                _entry = {
                    "file_name": file,
                    "sha1": _sha1,
                    "relative_path": _relative_path
                }

                unstaged_relative_paths.append(_relative_path)

                """
                If this is the first state, just add all files.
                """
                if self.__has_no_states:
                    unstaged_files["added"].append(_entry)
                    unstaged_files["count"] += 1
                    continue

                """
                Assuming this is not the first state, we need to check if
                the file is added, removed, modified or intact.
                """
                if self.__active_state.has_file(_sha1, _relative_path):
                    unstaged_files["intact"].append(_entry)
                elif orig := self.__active_state.has_relative_path(_relative_path):
                    if purpose == 1:
                        _sha1 = orig["sha1"]
                    unstaged_files["modified"].append({
                        "file_name": file,
                        "sha1": _sha1,
                        "relative_path": _relative_path
                    })
                    unstaged_files["count"] += 1
                    print(f"{_relative_path}is modified")
                else:
                    unstaged_files["added"].append(_entry)
                    unstaged_files["count"] += 1

                unstaged_relative_paths.append(_relative_path)

        if not self.__has_no_states:
            for file in list(active_state_files["added"].values()) + \
                        list(active_state_files["modified"].values()) + \
                        list(active_state_files["intact"].values()):
                for relative_path in file["relative_paths"]:
                    if relative_path not in unstaged_relative_paths:
                        unstaged_files["removed"].append({
                            "file_name": file["file_name"],
                            "sha1": file["sha1"],
                            "relative_path": relative_path
                        })
                        unstaged_files["count"] += 1

        return unstaged_files

    def commit(self, message: str, ignore: list = None) -> dict:
        """
        Commit the current state. This is a wrapper around the commit method
        of the FVSState class. A wrapper is used to store the state message
        and process staged files before committing.
        ...
        Raises:
            FVSEmptyCommitMessage: If the message is empty.
            FVSNothingToCommit: If there are no unstaged files.
        """
        if message in [None, ""]:
            raise FVSEmptyCommitMessage()

        unstaged_files = self.get_unstaged_files(ignore)
        if unstaged_files["count"] == 0:
            raise FVSNothingToCommit()

        # Create a new state
        state = FVSState(self)
        state.commit(message, unstaged_files)
        self.__states[state.state_id] = {
            "message": message,
            "timestamp": time.time()
        }
        self.__active_state = state
        self.__update_repo()
        return {
            "state_id": state.state_id,
            "message": message,
            "timestamp": time.time(),
            "added": len(unstaged_files["added"]),
            "removed": len(unstaged_files["removed"]),
            "modified": len(unstaged_files["modified"]),
            "intact": len(unstaged_files["intact"])
        }

    def delete_state(self, state_id: int, update_repo: bool = True):
        """
        Delete a state and all its subsequent states.
        ...
        Raises:
            FVSStateZeroNotDeletable: If the state_id is 0.
            FVSStateNotFound: If the state doesn't exist.
        """
        if int(state_id) == 0:
            raise FVSStateZeroNotDeletable()

        if int(state_id) not in self.__states:
            raise FVSStateNotFound(state_id)

        """
        Traveling in the future is probably something we don't want to do. So
        we will break references for subsequent states too.
        """
        for _state_id in [state_id] + self.__get_subsequent_state_ids(state_id):
            state = FVSState(self, _state_id)
            state.break_references()

            """
            If the state is the active state, we need to set the active state to
            the previous one.
            """
            if self.__active_state.state_id == _state_id:
                self.__active_state = FVSState(self, self.__get_prior_state_id(_state_id))

            """
            Delete the state from the states folder. It should be safer now as
            we already unreferenced the state from all its files.
            """
            self.__delete_state_folder(state)
            del self.__states[_state_id]

        if update_repo:
            self.__update_repo()

    def delete_active_state(self):
        """
        Delete the active state.
        """
        self.delete_state(self.__active_state.state_id)

    def restore_state(self, state_id: int, ignore: list = None):
        """
        Restore the state with the given id. This will remove all unstaged
        files and restore the given state, deleting any subsequent states.
        ...
        Raises:
            FVSStateNotFound: If the state doesn't exist.
            FVSNothingToRestore: If there are no unstaged files.
        """
        if int(state_id) not in self.__states.keys():
            raise FVSStateNotFound(state_id)

        self.__active_state = FVSState(self, state_id)
        subsequent_state_id = self.__get_subsequent_state_id(state_id)
        unstaged_files = self.get_unstaged_files(ignore, purpose=1)

        if unstaged_files["count"] == 0:
            raise FVSNothingToRestore()

        """
        If the given state has subsequent states, we need to delete them. The
        following call will start breaking references for the first subsequent
        state, FVSData will take care of the rest, physically deleting the
        files when the reference count reaches 0 (no state references).
        """
        if subsequent_state_id is not None:
            self.delete_state(subsequent_state_id, False)

        """
        Here we restore the situation to the given state, removing all
        unstaged files.
        """
        fvs_data = FVSData(self)

        for file in unstaged_files["added"]:
            _file_path = os.path.join(self.__repo_path, file["relative_path"])
            if os.path.isdir(_file_path):
                shutil.rmtree(_file_path)
            else:
                os.remove(_file_path)

        for file in unstaged_files["modified"]:
            internal_path = fvs_data.get_int_path(file["file_name"])
            FVSFile(self, file["file_name"], file["sha1"], [file["relative_path"]]).restore(internal_path)

        for file in unstaged_files["removed"]:
            internal_path = fvs_data.get_file_location(file["sha1"])
            FVSFile(self, file["file_name"], file["sha1"], [file["relative_path"]]).restore(internal_path)

        self.__update_repo()

    @staticmethod
    def __delete_state_folder(state: FVSState):
        """
        Delete the state folder with the given id.
        """
        shutil.rmtree(state.state_path)

    def is_valid_state(self, state_id: int) -> bool:
        """
        Check if the state with the given id is valid.
        ...
        Raises:
            FVSStateNotFound: If the state with the given id does not exist.
            FVSMissingStateIndex: If the state with the given id is missing
            FVSEmptyStateIndex: If the state with the given id is empty.
        """
        index_path = os.path.join(self.get_state_path(state_id), "files.json")

        if not os.path.exists(self.get_state_path(state_id)):
            raise FVSStateNotFound(state_id)

        if not os.path.exists(index_path):
            raise FVSMissingStateIndex(state_id)

        with open(index_path, "r") as f:
            index = orjson.loads(f.read())

        if not index:
            raise FVSEmptyStateIndex(state_id)

        return True

    def __get_prior_state_id(self, state_id: int) -> int:
        """
        Get the id of the prior state.
        ...
        Raises:
            FVSStateNotFound: If the state with the given id does not exist.
        """
        if int(state_id) not in self.__states.keys():
            raise FVSStateNotFound(state_id)

        for key in self.__states.keys():
            if key < state_id:
                return key

        return 0

    def __get_subsequent_state_ids(self, state_id: int) -> list:
        """
        Get the ids of the subsequent states.
        ...
        Raises:
            FVSStateNotFound: If the state with the given id does not exist.
        """
        if int(state_id) not in self.__states.keys():
            raise FVSStateNotFound(state_id)

        subsequent_states = []
        for key in self.__states.keys():
            if key > int(state_id):
                subsequent_states.append(key)

        return subsequent_states

    def __get_subsequent_state_id(self, state_id: int) -> int:
        """
        Get the id of the subsequent state.
        ...
        Raises:
            FVSStateNotFound: If the state with the given id does not exist.
        """
        if int(state_id) not in self.__states.keys():
            raise FVSStateNotFound(state_id)

        for key in self.__states.keys():
            if key > int(state_id):
                return key

        return 0

    def __get_relative_path(self, path: str) -> str:
        """
        Get the relative path of the given files.
        """
        repo_root = os.path.dirname(self.__repo_path)
        return os.path.relpath(path, self.__repo_path)

    def get_state_path(self, state_id: int) -> str:
        """
        Get the path of the state with the given id.
        """
        return os.path.join(self.__states_path, str(state_id))

    def __update_repo(self):
        """
        Update the repository configuration.
        """
        repo_conf = os.path.join(self.__repo_path, ".fvs/repo.json")
        with open(repo_conf, "wb") as f:
            self.__repo_conf["id"] = self.__active_state.state_id
            self.__repo_conf["states"] = self.__states
            f.write(orjson.dumps(self.__repo_conf, f, option=orjson.OPT_NON_STR_KEYS,))

        if self.__has_no_states:
            self.__has_no_states = False

    def new_state_path_by_id(self, state_id: int) -> str:
        """
        Get the path of the state with the given id.
        ...
        Raises:
            FVSStateAlreadyExists: If the state already exists.
        """
        state_path = os.path.join(self.__states_path, str(state_id))
        if os.path.exists(state_path):
            raise FVSStateAlreadyExists(state_id)
        os.makedirs(state_path)
        return state_path

    @property
    def repo_path(self) -> str:
        """
        Get the repository path.
        """
        return self.__repo_path

    @property
    def states_path(self) -> str:
        """
        Get the path of the states.
        """
        return self.__states_path

    @property
    def next_state_id(self) -> int:
        """
        Get the next state id.
        """
        if self.__has_no_states:
            return 0
        return list(self.__states)[-1] + 1

    @property
    def active_state_id(self) -> int:
        """
        Get the active state.
        """
        if self.__active_state is None:
            return None
        return self.__active_state.state_id

    @property
    def states(self) -> dict:
        """
        Get the list of states.
        """
        return self.__states
    
    @property
    def has_compression(self) -> bool:
        """
        Get the compression status.
        """
        return self.__use_compression

    @property
    def has_no_states(self) -> bool:
        """
        Get the no states status.
        """
        return self.__has_no_states
        


import os
import sys
import argparse
import datetime
import contextlib
from fvs.repo import FVSRepo
from fvs.exceptions import FVSNothingToCommit, FVSEmptyCommitMessage, FVSStateNotFound, FVSNothingToRestore

version = 'FVS 0.3.4'


def fvs_cli():
    parser = argparse.ArgumentParser(description='FVS')
    parser.add_argument('-v', '--version', action='version', version=version)

    subparsers = parser.add_subparsers(dest='command', help='sub-command help')

    init_parser = subparsers.add_parser("init", help="Create a new FVS repository")
    init_parser.add_argument('-i', '--ignore', help='patterns to ignore', action='append', default=[], required=False)
    init_parser.add_argument('-p', '--path', help='path to the repository', default=os.getcwd())
    init_parser.add_argument('-c', '--use-compression', help='use compression', action='store_true', default=False)

    commit_parser = subparsers.add_parser("commit", help="Commit changes to the repository")
    commit_parser.add_argument('-i', '--ignore', help='patterns to ignore', action='append', default=[], required=False)
    commit_parser.add_argument('-m', '--message', help='commit message', nargs='+', required=True)

    states_parser = subparsers.add_parser("states", help="List all states in the repository")

    restore_parser = subparsers.add_parser("restore", help="Restore a state from the repository")
    restore_parser.add_argument('-i', '--ignore', help='patterns to ignore', action='append', default=[], required=False)
    restore_parser.add_argument('-s', '--state-id', help='state id', required=True)

    args = parser.parse_args()

    if args.command == 'init':
        repo = FVSRepo(args.path, args.use_compression)

        with contextlib.suppress(FVSNothingToCommit):
            repo.commit("Init", args.ignore)

        sys.stdout.write("Initialized FVS repository in {}\n".format(args.path))
        sys.exit(0)

    elif args.command == 'commit':
        repo = FVSRepo(os.getcwd())
        message = ' '.join(args.message)

        try:
            sys.stdout.write("Committing...\n")
            res = repo.commit(message, args.ignore)
            sep = "-" * 10
            sys.stdout.write("\nCommitted state {}\nMessage: {}\nDate: {}\n{}\nAdded files: {}\nRemoved files: {}\nModified files: {}\nIntact files: {}\n".format(
                res['state_id'],
                res['message'],
                datetime.datetime.fromtimestamp(res['timestamp']).strftime('%Y-%m-%d %H:%M:%S'),
                sep,
                res["added"],
                res["removed"],
                res["modified"],
                res["intact"]
            ))
            sys.exit(0)
        except FVSNothingToCommit:
            sys.stderr.write("Nothing to commit\n")
            sys.exit(1)
        except FVSEmptyCommitMessage:
            sys.stderr.write("Empty commit message\n")
            sys.exit(1)

    elif args.command == 'states':
        repo = FVSRepo(os.getcwd())

        if len(repo.states) == 0:
            sys.stdout.write("No states\n")
            sys.exit(0)

        for k, v in repo.states.items():
            sym = "\033[32m➔" if k == repo.active_state_id else "-"
            sys.stdout.write(
                "{} ({}): {}\n\t{}\n\n\033[0m".format(
                    sym, k, v["message"], datetime.datetime.fromtimestamp(v["timestamp"]))
            )

        sys.exit(0)

    elif args.command == 'restore':
        repo = FVSRepo(os.getcwd())
        try:
            repo.restore_state(args.state_id, args.ignore)
            sys.stdout.write("Restored state\n")
            sys.exit(0)
        except FVSStateNotFound:
            sys.stderr.write("State {} not found\n".format(args.state_id))
            sys.exit(1)
        except FVSNothingToRestore:
            sys.stderr.write("Nothing to restore from state {}\n".format(args.state_id))
            sys.exit(1)

    elif args.command == 'active':
        repo = FVSRepo(os.getcwd())
        if repo.active_state_id in [-1, None]:
            sys.stdout.write("No active state\n")
            sys.exit(0)
        sys.stdout.write("Active state is {}\n".format(repo.active_state_id))
        sys.exit(0)

    else:
        parser.print_help()
        sys.exit(1)


if __name__ == '__main__':
    fvs_cli()
    sys.exit(0)



import os
import shutil
import tarfile
import logging

logger = logging.getLogger("fvs.file")


# noinspection DuplicatedCode
class FVSFile:

    def __init__(self, repo: 'FVSRepo', file_name: str, sha1: str, relative_paths: list):
        self.__repo = repo
        self.__file_name = file_name
        self.__sha1 = sha1
        self.__relative_paths = relative_paths

    def is_equal(self, file: 'FVSFile') -> bool:
        return self.__sha1 == file.get_sha1()

    def as_dict(self) -> dict:
        return {
            "file_name": self.__file_name,
            "sha1": self.__sha1,
            "relative_paths": self.__relative_paths
        }

    def copy_to(self, dest: str, use_sha1_as_name: bool = True):
        """
        This method opy the file to the given destination. Despite it looks 
        flexible, it is meant to be used only by FVSData to copy files to 
        the appropriate data location, for this reason use_sha1_as_name is 
        set to True by default (data files must be stored with their sha1 
        hash as name to avoid name collisions). This method use copy2 to
        copy the file, so it will preserve the file metadata.
        """

        if self.__repo.has_compression:
            self.__compress_copy_to(dest, use_sha1_as_name)
            return

        if use_sha1_as_name:
            _dest = os.path.join(dest, self.__sha1)
            _name = self.__sha1
        else:
            _dest = os.path.join(dest, self.__file_name)
            _name = self.__file_name

        """
        There never should be a file with the same name and FVSData should
        already check for duplicates. Anyway, we will check for it here
        just in case.
        """
        if os.path.islink(_dest) or os.path.exists(_dest):
            logger.debug(f"File {self.__sha1} already exists in {dest}.")
            return

        """
        We will move only the first relative path as the file is supposed to
        be the same in all relative paths.
        """
        logger.debug(f"Copying file {_name} to {dest}")
        shutil.copy2(
            os.path.join(self.__repo.repo_path, self.__relative_paths[0]),
            _dest,
            follow_symlinks=False
        )

    def remove(self, path: str, use_sha1_as_name: bool = True):
        """
        This method will remove the file from the internal data directory.
        """
        if use_sha1_as_name:
            file_path = os.path.join(path, self.__sha1)
        else:
            file_path = os.path.join(path, self.__file_name)

        if os.path.exists(file_path):
            logger.debug(f"removing file {self.__file_name} from {path}")
            os.remove(file_path)
        else:
            logger.debug(f"file {self.__file_name} does not exist, data catalog may be corrupted.")

    def restore(self, internal_path: str):
        """
        This method will restore the file, copying from the internal data
        directory to the repo, renaming it to the original name.
        """
        if self.__repo.has_compression:
            self.__compress_restore(internal_path)
            return
            
        file_path = os.path.join(internal_path, self.__sha1)
        if not os.path.exists(file_path):
            logger.debug(f"file {self.__file_name} does not exist, data catalog may be corrupted.")
            return

        for relative_path in self.__relative_paths:
            dir_name = os.path.dirname(os.path.join(self.__repo.repo_path, relative_path))
            logger.debug(f"restoring file {self.__file_name}")
            os.makedirs(dir_name, exist_ok=True)
            shutil.copy2(
                file_path,
                os.path.join(self.__repo.repo_path, relative_path),
                follow_symlinks=False
            )
    
    def __compress_copy_to(self, dest: str, use_sha1_as_name: bool = True):
        """
        This method will copy the file to the given destination, compressing it.
        """
        if use_sha1_as_name:
            _dest = os.path.join(dest, self.__sha1)
            _name = self.__sha1
        else:
            _dest = os.path.join(dest, self.__file_name)
            _name = self.__file_name

        """
        There never should be a file with the same name and FVSData should
        already check for duplicates. Anyway, we will check for it here
        just in case.
        """
        if os.path.islink(_dest) or os.path.exists(_dest):
            logger.debug(f"File {self.__sha1} already exists in {dest}.")
            return

        """
        We will move only the first relative path as the file is supposed to
        be the same in all relative paths.
        """
        logger.debug(f"Compressing file {_name} to {dest}")
        with tarfile.open(_dest, "w:gz") as tar:
            tar.add(
                os.path.join(self.__repo.repo_path, self.__relative_paths[0]),
                arcname=_name
            )
    
    def __compress_restore(self, internal_path: str):
        """
        This method will restore the file, decompressing it and copying it
        to the repo.
        """
        file_path = os.path.join(internal_path, self.__sha1)
        if not os.path.exists(file_path):
            logger.debug(f"file {self.__file_name} does not exist, data catalog may be corrupted.")
            return

        for relative_path in self.__relative_paths:
            full_rel_path = os.path.join(self.__repo.repo_path, relative_path)
            dir_name = os.path.dirname(full_rel_path)
            logger.debug(f"restoring file {self.__file_name}")
            os.makedirs(dir_name, exist_ok=True)
            with tarfile.open(file_path, "r:gz") as tar:
                tar.extractall(dir_name)
                os.rename(
                    os.path.join(dir_name, self.__sha1),
                    full_rel_path
                )

    @property
    def file_name(self) -> str:
        return self.__file_name

    @property
    def sha1(self) -> str:
        return self.__sha1

    @property
    def relative_paths(self) -> list:
        return self.__relative_paths



import fnmatch
import logging

logger = logging.getLogger("fvs.pattern")


class FVSPattern:

    @staticmethod
    def match(patterns: list, file_name: str) -> bool:
        """
        This method will check if the file_name matches the pattern. It is
        currently just a wrapper around fnmatch.fnmatch. Here just for better
        code readability and future improvements.
        """
        for pattern in patterns:
            if fnmatch.fnmatch(file_name, pattern):
                logger.debug(f"One pattern match: {file_name} matches {pattern}")
                return True
        return False



import os
import orjson
import logging

from fvs.exceptions import FVSCallerWrongClass, FVSEmptyCommitMessage, FVSWrongUnstagedDict, \
    FVSStateNotFound, FVSCommittingToExistingState, FVSUnsupportedKey
from fvs.data import FVSData
from fvs.file import FVSFile
from fvs.utils import FVSUtils

logger = logging.getLogger("fvs.state")


class FVSState:
    __files: dict = None
    __state_id: int = None
    __state_path: str = None

    def __init__(self, repo: 'FVSRepo', state_id: int = None):
        self.__repo = repo
        self.__files = {"count": 0, "added": {}, "modified": {}, "removed": {}, "intact": {}}

        if state_id is not None:
            self.__load_state(state_id)

    @classmethod
    def load(cls, repo: 'FVSRepo', state_id: int):
        """
        This constructor will return a new FVSState object for the given
        state_id.
        """
        return cls(repo, state_id)

    def __load_state(self, state_id: int):
        """
        This method will load a state from the repository.
        """
        self.__state_id = state_id
        self.__state_path = os.path.join(self.__repo.states_path, str(state_id))
        if not os.path.exists(self.__state_path):
            raise FVSStateNotFound(state_id)

        with open(os.path.join(self.__state_path, "files.json"), "r") as f:
            self.__files = orjson.loads(f.read())
        
    def commit(
            self,
            message: str,
            unstaged_files: dict
    ):
        """
        States are supposed to be committed only from FVSRepo and only on first
        initialization. So here we will check if the caller is the expected
        class.
        """
        if FVSUtils.get_caller_class_name() != "FVSRepo":
            raise FVSCallerWrongClass("FVSRepo")

        """
        For the same reason above, we will check if the state were already
        initialized to avoid unwanted commits.
        """
        if self.__is_initialized():
            raise FVSCommittingToExistingState()

        """
        As a rule in FVS, the commit message should not be empty. We don't
        want untraceable commits.
        """
        if message in [None, ""]:
            raise FVSEmptyCommitMessage()

        """
        To avoid further investigation and multiple checks, we will check
        for the unstaged files dict structure. It must contain the following
        keys: added, modified, removed.
        """
        if False in [
            unstaged_files.get("count"),
            unstaged_files.get("added"),
            unstaged_files.get("modified"),
            unstaged_files.get("removed"),
            unstaged_files.get("intact")
        ]:
            raise FVSWrongUnstagedDict()

        """
        Set the state id with the next state id available in the repository.
        """
        self.__state_id = self.__repo.next_state_id
        self.__files["count"] = unstaged_files["count"]

        """
        Instantiate the FVSData class and start collecting the files.
        """
        fvs_data = FVSData(self.__repo, self)

        for _file in unstaged_files["added"]:
            fvs_data.add_file(FVSFile(self.__repo, _file["file_name"], _file["sha1"], [_file["relative_path"]]))
            if _file["sha1"] in self.__files["added"]:
                self.__files["added"][_file["sha1"]]["relative_paths"] += [_file["relative_path"]]
            else:
                self.__files["added"][_file["sha1"]] = {
                    "file_name": _file["file_name"],
                    "sha1": _file["sha1"],
                    "relative_paths": [_file["relative_path"]],
                }

        for _file in unstaged_files["modified"]:
            fvs_data.add_file(FVSFile(self.__repo, _file["file_name"], _file["sha1"], [_file["relative_path"]]))
            if _file["sha1"] in self.__files["modified"]:
                self.__files["modified"][_file["sha1"]]["relative_paths"] += [_file["relative_path"]]
            else:
                self.__files["modified"][_file["sha1"]] = {
                    "file_name": _file["file_name"],
                    "sha1": _file["sha1"],
                    "relative_paths": [_file["relative_path"]],
                }

        for _file in unstaged_files["removed"]:
            fvs_data.delete_file(FVSFile(self.__repo, _file["file_name"], _file["sha1"], [_file["relative_path"]]))
            if _file["sha1"] in self.__files["removed"]:
                self.__files["removed"][_file["sha1"]]["relative_paths"] += [_file["relative_path"]]
            else:
                self.__files["removed"][_file["sha1"]] = {
                    "file_name": _file["file_name"],
                    "sha1": _file["sha1"],
                    "relative_paths": [_file["relative_path"]],
                }

        for _file in unstaged_files["intact"]:
            if _file["sha1"] in self.__files["intact"]:
                self.__files["intact"][_file["sha1"]]["relative_paths"] += [_file["relative_path"]]
            else:
                self.__files["intact"][_file["sha1"]] = {
                    "file_name": _file["file_name"],
                    "sha1": _file["sha1"],
                    "relative_paths": [_file["relative_path"]],
                }

        fvs_data.complete_transaction()
        self.__save_state()

    def break_references(self):
        """
        This method ask to FVSData to remove the reference to the state for 
        all the files, it will also physical delete the file if it has
        no other referenced states.
        """
        if FVSUtils.get_caller_class_name() != "FVSRepo":
            raise FVSCallerWrongClass("FVSRepo")

        fvs_data = FVSData(self.__repo, self)

        for _file in self.__files["added"].values():
            fvs_data.delete_file(FVSFile(self.__repo, _file["file_name"], _file["sha1"], _file["relative_paths"]))

        for _file in self.__files["modified"].values():
            fvs_data.delete_file(FVSFile(self.__repo, _file["file_name"], _file["sha1"], _file["relative_paths"]))

        fvs_data.complete_transaction()

    def has_file(self, sha1: str, relative_path: str) -> bool:
        """
        This method will check if the state has the given file.
        """
        if sha1 in self.__files["added"]:
            if relative_path in self.__files["added"][sha1]["relative_paths"]:
                return True
        if sha1 in self.__files["modified"]:
            if relative_path in self.__files["modified"][sha1]["relative_paths"]:
                return True
        if sha1 in self.__files["intact"]:
            if relative_path in self.__files["intact"][sha1]["relative_paths"]:
                return True
                
        return False

    def __save_state(self):
        """
        This method will save the state to the repository.
        """
        state_path = self.__repo.new_state_path_by_id(self.__state_id)
        with open(os.path.join(state_path, "files.json"), "wb") as f:
            f.write(orjson.dumps(self.__files, f, option=orjson.OPT_NON_STR_KEYS,))

    def __is_initialized(self) -> bool:
        """
        This method will check if the state is initialized.
        """
        try:
            self.__repo.is_valid_state(self.__repo.next_state_id)
        except FVSStateNotFound:
            return False
        return True

    def has_relative_path(self, relative_path: str, key: str = "any") -> bool:
        """
        This method will return the entry from the state files which
        corresponds to the given file name. The 'any' key will check in
        added, modified and intact files.
        """
        supported_keys = ["any", "added", "modified", "intact"]
        if key not in supported_keys:
            raise FVSUnsupportedKey(supported_keys)

        if key == "any":
            for file in self.__files["added"].values():
                if relative_path in file["relative_paths"]:
                    return file
            for file in self.__files["modified"].values():
                if relative_path in file["relative_paths"]:
                    return file
            for file in self.__files["intact"].values():
                if relative_path in file["relative_paths"]:
                    return file
        else:
            for _file in self.__files[key].values():
                if relative_path in _file["relative_paths"]:
                    return _file
        return None

    @property
    def files(self) -> dict:
        """
        This method will return the files in the state.
        """
        return self.__files

    @property
    def state_id(self) -> int:
        """
        This method will return the state id.
        """
        return self.__state_id

    @property
    def state_path(self) -> str:
        """
        This method will return the state path.
        """
        return self.__state_path



import os
import inspect
import hashlib
from typing import Union


class FVSUtils:

    @staticmethod
    def get_caller_class_name() -> str:
        """
        Get the name of the caller class.
        """
        stack = inspect.stack()[2]
        caller = stack[0].f_locals.get('self').__class__.__name__
        return caller

    @staticmethod
    def get_sha1_hash(path: str, block_size: int = 2 ** 20) -> Union[str, None]:
        """
        Get the sha1 hash of the given file. It will use name+content
        to avoid empty files.
        """
        sha1_temp = hashlib.sha1()
        file_name = os.path.basename(path)
        try:
            with open(path, "rb") as f:
                while True:
                    buffer = f.read(block_size)
                    if not buffer:
                        break
                    sha1_temp.update(buffer)
            sha1_temp.update(file_name.encode())
            return sha1_temp.hexdigest()
        except (FileNotFoundError, PermissionError, IsADirectoryError):
            return None



class FVSException(Exception):
    """
    Base class for all FVS exceptions.
    """

    def __init__(self, message: str = "An unknown error occurred."):
        super().__init__(message)


class FVSStateNotFound(FVSException):
    """
    Exception raised when a state is not found.
    """

    def __init__(self, state_id: int):
        super().__init__("No state found for ID: {}".format(state_id))


class FVSEmptyStateIndex(FVSException):
    """
    Exception raised when the state index is empty.
    """

    def __init__(self, state_id: int):
        super().__init__("Index is empty for state ID: {}".format(state_id))


class FVSMissingStateIndex(FVSException):
    """
    Exception raised when a state index is not found.
    """

    def __init__(self, state_id: int):
        super().__init__("State index not found for state with ID: {}".format(state_id))


class FVSFileNotFound(FVSException):
    """
    Exception raised when a file is not found.
    """

    def __init__(self):
        super().__init__("File not found in the state.")


class FVSFileAlreadyExists(FVSException):
    """
    Exception raised when a file already exists.
    """

    def __init__(self):
        super().__init__("File already exists.")


class FVSNothingToCommit(FVSException):
    """
    Exception raised when there is nothing to commit.
    """

    def __init__(self):
        super().__init__("Nothing to commit.")


class FVSEmptyCommitMessage(FVSException):
    """
    Exception raised when the commit message is empty.
    """

    def __init__(self):
        super().__init__("Commit message is empty.")


class FVSCallerWrongClass(FVSException):
    """
    Exception raised when the caller is not the expected class.
    """

    def __init__(self, cls: str = "FVSRepo"):
        super().__init__("Caller is not the expected class: {}".format(cls))


class FVSWrongUnstagedDict(FVSException):
    """
    Exception raised when the unstaged_files dict is not correct.
    """

    def __init__(self):
        super().__init__("The unstaged_files dict is not correct (following\
keys are required: added, modified, removed, intact).")


class FVSUnsupportedKey(FVSException):
    """
    Exception raised when the has_relative_path method was requested
    with a wrong key.
    """

    def __init__(self, supported_keys: list):
        super().__init__("The has_relative_path method was requested with a \
wrong key, the following are corrected: {}.".format(supported_keys))


class FVSCommittingToExistingState(FVSException):
    """
    Exception raised when committing to an existing state.
    """

    def __init__(self):
        super().__init__("You were trying to commit to an existing state. \
This is not allowed and states are not meant to be altered \
once the first commit has been made.")


class FVSDataHasNoState(FVSException):
    """
    Exception raised when the state data has no state.
    """

    def __init__(self):
        super().__init__("FVSData was initialized without a FVSState, \
but a FVSState was expected to perform transactions.")


class VFSTransactionAlreadyStarted(FVSException):
    """
    Exception raised when a transaction is already started.
    """

    def __init__(self):
        super().__init__("A different kind of transaction is already started.")


class FVSWrongSortBy(FVSException):
    """
    Exception raised when a sort_by parameter is not correct.
    """

    def __init__(self, allowed_sort_by: list):
        super().__init__("The sort_by parameter is not correct. \
It should be one of the following: {}".format(allowed_sort_by))


class FVSNothingToRestore(FVSException):
    """
    Exception raised when there is nothing to restore.
    """

    def __init__(self):
        super().__init__("Nothing to restore.")


class FVSStateZeroNotDeletable(FVSException):
    """
    Exception raised when a state with ID 0 is not deletable.
    """

    def __init__(self):
        super().__init__("State with ID 0 is not deletable.")


class FVSStateAlreadyExists(FVSException):
    """
    Exception raised when a state already exists.
    """

    def __init__(self, state_id: int):
        super().__init__("State already exists with ID: {}".format(state_id))



import os
import orjson
import logging
from typing import Union

from fvs.exceptions import FVSDataHasNoState, VFSTransactionAlreadyStarted

logger = logging.getLogger("fvs.data")


# noinspection PyTypeChecker
class FVSData:
    __data_conf: dict = None
    __data_conf_path: str = None
    __data_int_paths: list = [
        "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m",
        "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z",
        "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "-"
    ]
    __state: 'FVSState' = None
    __transaction: list = None
    __transaction_type: int = None  # 0: add, 1: remove

    def __init__(self, repo: 'FVSRepo', state: 'FVSState' = None):
        """
        Initialize the FVSData.
        """
        self.__data_path = os.path.join(repo.repo_path, ".fvs/data")
        self.__state = state
        self.__update_fvs_path()
        self.__load_config()

    def __update_fvs_path(self):
        """
        Update the structure of path data/ in the FVS repository. This also
        performs some checks to ensure that the data/ directory is valid,
        fixing it if necessary.
        """
        self.__data_conf_path = os.path.join(self.__data_path, "data.json")

        """
        This check if the data/ directory exists. If not, create it.
        """
        if not os.path.exists(self.__data_path):
            os.makedirs(os.path.join(self.__data_path, dir))

        """
        This check if internal paths (__data_int_paths) exist. If not, 
        create them.
        """
        for int_path in self.__data_int_paths:
            if not os.path.exists(os.path.join(self.__data_path, int_path)):
                os.makedirs(os.path.join(self.__data_path, int_path))

        """
        Check if the data.json file exists. If not, create it and write the
        default configuration.
        """
        if not os.path.exists(self.__data_conf_path):
            with open(self.__data_conf_path, "wb") as f:
                self.__data_conf = {}
                f.write(orjson.dumps(self.__data_conf, f, option=orjson.OPT_NON_STR_KEYS,))

    def __load_config(self):
        """
        Load the data configuration from the data/data.json file.
        """
        with open(self.__data_conf_path, "r") as f:
            self.__data_conf = orjson.loads(f.read())

    def __save_config(self):
        """
        Save the data configuration to the data/data.json file.
        """
        with open(self.__data_conf_path, "wb") as f:
            f.write(orjson.dumps(self.__data_conf, f, option=orjson.OPT_NON_STR_KEYS,))

    def complete_transaction(self):
        """
        Complete the transaction duplicating the files in the proper internal
        data path. It also saves the configuration to the data/data.json file.
        """
        if self.__transaction is None:
            return  # it's safe to ignore this call, the state is probably only removing files

        for file in self.__transaction:
            _int_path = self.get_int_path(file.file_name)

            if self.__transaction_type == 0:
                file.copy_to(_int_path)
            elif self.__transaction_type == 1:
                file.remove(_int_path)

        self.__save_config()

    def get_int_path(self, file_name: str) -> str:
        """
        This simple method determines the internal path of a file based
        on the first letter of the file name. Every file starting with
        a special character will be placed in the "-" directory.
        """
        first_letter = file_name[0].lower()
        int_path = "-"
        if first_letter in self.__data_int_paths:
            int_path = first_letter

        return os.path.join(self.__data_path, int_path)

    def __set_transaction_type(self, type_id: int):
        if self.__transaction is None:
            self.__transaction = []
            if self.__transaction_type is None:
                self.__transaction_type = type_id
            elif self.__transaction_type != type_id:
                raise VFSTransactionAlreadyStarted()

    def add_file(self, file: 'FVSFile'):
        """
        This method add a file to the catalog and append it to the
        transaction list. Files already in the catalog will be updated 
        listing the new state for the deduplication. A FVSFile object
        is needed for the 'file' parameter.
        ...
        Raises:
            FVSDataHasNoState: if the state is not set.
        """
        if not self.__state:
            raise FVSDataHasNoState()

        self.__set_transaction_type(0)

        if file.sha1 not in self.__data_conf.keys():
            logger.debug(f"Adding file {file.file_name} to data catalog.")
            self.__data_conf[file.sha1] = {
                "file_name": file.file_name,
                "sha1": file.sha1,
                "states": {str(self.__state.state_id): 1}
            }
            self.__transaction.append(file)

        elif str(self.__state.state_id) not in self.__data_conf[file.sha1]["states"]:
            logger.debug(f"Adding state {self.__state.state_id} to file {file.file_name} in data catalog.")
            self.__data_conf[file.sha1]["states"][str(self.__state.state_id)] = 1
            self.__transaction.append(file)

        else:
            logger.debug(f"File {file.file_name} already in data catalog.")
            self.__data_conf[file.sha1]["states"][str(self.__state.state_id)] += 1

    def delete_file(self, file: 'FVSFile', state_id: int = None):
        """
        This method delete a state for a file in the catalog, it will
        remove the file entry if the state is the last one. If state_id
        is not set, the one defined in the FVSFile object will be used, 
        assuming it was the intended state.
        ...
        Raises:
            FVSDataHasNoState: if the state is not set.
        """
        if state_id is None and not self.__state:
            raise FVSDataHasNoState()

        if state_id is None:
            state_id = self.__state.state_id

        self.__set_transaction_type(1)

        if file.sha1 in self.__data_conf.keys():
            if str(state_id) in self.__data_conf[file.sha1]["states"]:
                logger.debug(f"Unlinking state {state_id} from file {file.file_name} in data catalog.")
                self.__data_conf[file.sha1]["states"][str(state_id)] -= 1

                if self.__data_conf[file.sha1]["states"][str(state_id)] == 0:
                    logger.debug(
                        f"{file.file_name} reached 0 for state {state_id}. Removing state reference.")
                    del self.__data_conf[file.sha1]["states"][str(state_id)]
                    self.__transaction.append(file)
            
                if len(self.__data_conf[file.sha1]["states"]) == 0:
                    logger.debug(f"{file.file_name} reached 0 for all states. Removing from data catalog.")
                    del self.__data_conf[file.sha1]
                    self.__transaction.append(file)

            else:
                logger.debug(f"File {file.file_name} has no state {self.__state.state_id} referenced. Ignoring.")
        else:
            logger.debug(f"File {file.file_name} is not in data catalog. Ignoring.")

    def get_file_location(self, sha1: str) -> Union[str, None]:
        """
        This method returns the location of a file in the data catalog.
        """
        if sha1 in self.__data_conf.keys():
            file_name = self.__data_conf[sha1]["file_name"]
            return self.get_int_path(file_name)
        else:
            logging.debug(f"File {sha1} is not in data catalog.")
            return None



from setuptools import setup

setup(
    name='FVS',
    version='0.3.4',
    packages=['fvs'],
    url='https://github.com/mirkobrombin/FVS',
    license='MIT',
    author='Mirko Brombin',
    author_email='send@mirko.pm',
    description='File Versioning System with hash comparison, deduplication and data storage to create unlinked '
                'states that can be deleted ',
    entry_points={
        'console_scripts': [
            'fvs=fvs.cli:fvs_cli'
        ]
    },
    install_requires=[
        'orjson'
    ]
)
